(* $Id: godi_plan.mli 64 2003-08-27 22:05:35Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Analyze dependencies and plan the build process *)

open Godi_types
open Godi_exec

exception Mission_impossible of string
  (* The string arguments describes the most likely reason. *)

val get_requirements : 
      available:package list ->       (* Which packages are available *)
      selected:package list ->        (* Which packages are already selected *)
      unit ->
	package list

      (* Checks which packages must be also selected to match the requirements
       * of the selected packages.
       * The returned package list is the list of missing requirements that
       * have been found in the list of available packages (including the
       * already selected packages). The returned list always contains
       * all members of [selected], plus the required members of
       * [available]. If there is no solution for this problem, the
       * exception Mission_impossible will be raised.
       *
       * All packages can be source packages, binary packages or installed
       * packages.
       *
       * SOURCE EXAMPLE:
       *
       * available: The list of installed packages + The list of non-installed
       *   source packages
       * selected: The list of installed packages + The packages the user
       *   wishes to build
       *
       * BINARY EXAMPLE:
       *
       * available: The list of installed packages + The list of non-installed
       *   binary packages
       * selected: The list of installed packages + The packages the user
       *   wishes to add
       *
       * It is possible to mix source and binary packages. If there are
       * several versions for a package, the most recent version is selected.
       * Installed packages are preferred over binaray packages, which
       * are preferred over source packages.
       *)

val get_dependent_packages :
      active:package list ->      (* Usually: Which packages are installed *)
      selected:package list ->    (* The selected packages *)
      unit ->
	package list

      (* Returns the active packages that are dependent on selected 
       * packages. The resulting list is a subset of the list of active
       * packages only. 
       *
       * The function does not check whether the version conditions of
       * dependencies match, nor whether any requirement is missing.
       * Only dependencies, not build dependencies are analyzed.
       * 
       * The active packages should not contain several packages with the
       * same name; otherwise strange results may be returned.
       *
       * Example:
       * - active = all installed packages
       * - selected = selected source packages
       *)

val sort_packages : package list -> package list
      (* Sorts the packages topologically; the deepest ancestor comes
       * first. Duplicates are removed.
       *)

val update_plan : 
      iteration:int ->
      package list -> 
	int
      (* Plans the build process, and updates the user_status field of the
       * packages.
       *
       * The idea of this function is to feed all package objects as input
       * into it, and let this function decide what to do by updating the
       * user_status fields. What can happen:
       *
       * - Installed packages can be set to `Remove_dep, because they
       *   need to be rebuilt, and this implies the installed package
       *   must be removed first
       * - Installed packages can be set to `Remove_dep because of
       *   dependencies on `Remove packages
       * - Source packages can be set to `Build_dep, because they
       *   need to be rebuilt as result of dependency analysis
       *
       * `Remove_dep and `Build_dep have an iteration number as argument;
       * this is set to the passed value of ~iteration.
       *
       * The function returns the number of packages that get `Build_dep
       * or `Remove_dep for the first time. If this number is positive, there
       * are packages that still need to be confirmed, and the user dialog 
       * to select packages should be shown again (next iteration).
       *)
