(* $Id: godi_query.ml 265 2004-02-11 00:41:09Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_versions
open Printf


let location loc =
  "File " ^ loc.Godi_make.loc_filename ^ ", line " ^
  string_of_int loc.Godi_make.loc_linenumber
;;


let error_msg exc =
  match exc with
      Failure s -> s
    | Command_failure(cmd,args,code) ->
	( match code with
	      WEXITED n ->
		sprintf "Command fails with code %d: %s" n cmd
	    | WSIGNALED n ->
		sprintf "Command terminated because of signal %d: %s" n cmd
	    | WSTOPPED _ ->
		sprintf "Command stopped: %s" cmd
	)
    | Command_not_found cmd ->
	sprintf "Command not found: %s" cmd
    | Command_discontinued ->
	"Command aborted"
    | General_error ->
	"Error"
    | Godi_make.Parse_error(loc, msg) ->
	sprintf "%s [Makefile parsing]: %s" (location loc) msg
    | Godi_make.Eval_error(loc, msg) ->
	sprintf "%s [Makefile evaluation]: %s" (location loc) msg
    | Godi_make.File_not_found file ->
	sprintf "File not found: %s" file
    | Godi_match.Bad_glob_expr s ->
	sprintf "Bad glob expression: %s" s
    | Godi_match.Bad_regular_expr s ->
	sprintf "Bad regular expression: %s" s
    | Godi_match.Unsupported_expr s ->
	sprintf "Unsupported glob/regular expression: %s" s
    | Godi_match.Bad_subst_expr ->
	"Bad substitution"
    | Godi_match.No_percent_symbol ->
	"Missing '%' as substitution identifier"
    | _ ->
	Printexc.to_string exc
;;



let to_prot protocol_option msg =
  match protocol_option with
      None -> ()
    | Some p -> collect p msg
;;


let tws_re = Str.regexp "[ \t\r\n]*$";;

let strip_trailing_ws s =
  Str.global_replace tws_re "" s
;;


let lazy_etc_dir =
  lazy (
    let b = Buffer.create 100 in
    call_cmd ~stdout:(`Buffer b) ("godi_confdir", [| |]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let lazy_godi_conf =
  lazy (
    let etc_dir = Lazy.force lazy_etc_dir in
    Filename.concat etc_dir "godi.conf"
  )
;;


let lazy_localbase =
  lazy (
    let godi_conf = Lazy.force lazy_godi_conf in
    let b = Buffer.create 100 in
    call_cmd
      ~stdout:(`Buffer b)
      ("godi_make",
       [| "-f"; godi_conf; "-V"; "LOCALBASE"; "all"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let lazy_pkgdbdir =
  lazy (
    let godi_conf = Lazy.force lazy_godi_conf in
    let b = Buffer.create 100 in
    call_cmd
      ~stdout:(`Buffer b)
      ("godi_make",
       [| "-f"; godi_conf; "-V"; "GODI_DBDIR"; "all"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let godi_conf() = Lazy.force lazy_godi_conf
;;


let localbase() = Lazy.force lazy_localbase
;;


let pkgdbdir() = Lazy.force lazy_pkgdbdir
;;


let buildbase() = Filename.concat (localbase()) "build" ;;


let get_tools_makefile() =
  try
    ignore(Sys.getenv "GODI_TOOLS_DEV");
    (* Assuming that cwd is "console" *)
    Sys.getcwd() ^ "/../godi-pkg/godi_tools.mk"
  with
      Not_found ->
	let bbase = buildbase() in
	let dir_mk = Filename.concat bbase "mk" in
	Filename.concat dir_mk "godi_tools.mk"
;;


let lazy_pkgtools_version =
  lazy (
    let b = Buffer.create 100 in
    let file = get_tools_makefile() in
    call_cmd
      ~stdout:(`Buffer b)
      ("godi_make",
       [| "-f"; file; "print-pkgtools-version"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let pkgtools_version() = Lazy.force lazy_pkgtools_version ;;



let read_file filename =
  (* Return the contents of [filename] *)
  let f = open_in filename in
  try
    let n = in_channel_length f in
    let s = String.create n in
    really_input f s 0 n;
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let read_optional_file filename =
  (* Returns the contents of [filename], if existing, and "" otherwise *)
  if Sys.file_exists filename then
    read_file filename
  else
    ""
;;


(***********************************************************************)
(* Source packages                                                     *)
(***********************************************************************)

let lines_re = Str.regexp "\n";;

let sloppy_lines_re = Str.regexp "[ \t]*\n[ \t]*";;

let words_re = Str.regexp "[ \t]+";;

let dep_re = Str.regexp "\\([^:]*\\):";;



let parse_dep_string s =
  if Str.string_match dep_re s 0 then (
    let cond_str = Str.matched_group 1 s in
    parse_dependency cond_str
  )
  else
    failwith ("Bad dependency: " ^ s)
;;


let parse_dep_strings s =
  let words = Str.split words_re s in
  List.map parse_dep_string words
;;


let classic_query_makefile ?protocol ?godi_conf ?(makefile="Makefile") dir variables =
  (* TODO: Write a quicker version that does not call "make" to get
   * the values of variables. A better method is to scan the Makefile,
   * and to interpret the definitions ourselves, and to fall back to
   * "make" only if we find conditional or computed definitions.
   *)
  let env =
    match godi_conf with
	None -> []
      | Some cf -> [ "GODI_CONF", cf ] in
  let b = Buffer.create 1000 in
  call_cmd
    ~cd:dir
    ~env
    ?stderr:protocol
    ~stdout:(`Buffer b)
    ("godi_make",
     [| "-f"; makefile;
        "show-vars";
	"VARNAMES=" ^ String.concat " " variables |]);
  let s = Buffer.contents b in
  assert(s <> "" && s.[ String.length s - 1 ] = '\n');
  (* Remove the trailing \n: *)
  let s' = String.sub s 0 (String.length s - 1) in
  (* Split into lines: *)
  let l = Str.split_delim lines_re s' in
  assert(List.length l = List.length variables);
  l
;;


let pcache = Godi_make.create_parser_cache();;

let modern_query_makefile ?protocol ?godi_conf:gcarg ?(makefile="Makefile") dir variables =

  let print_error msg =
    match protocol with
	None ->
	  prerr_string msg;
	  flush stderr
      | Some p ->
	  Godi_exec.collect p msg
  in

  let gc =
    match gcarg with
	None    -> godi_conf()
      | Some cf -> cf
  in

  let env =
    Array.concat [Unix.environment(); [| "GODI_CONF=" ^  gc |] ] in
  (* Latter variables override previous instances ! *)

  (* print_error ("Doing " ^ dir ^ "\n"); *)

  let sys_mk_path =
    [ Filename.concat (localbase()) "share/mk" ] in
  (* TODO: Remove when Godi_make correctly initializes itself *)

  (* Notes:
   * - There is always GODI_CONF in the environment
   * - PKGTOOLS_VERSION is also set
   * The result is that certain time-consuming actions need not to
   * be done when the Makefile is evaluated (running further external
   * programs)
   *)

  let ctx = Godi_make.create_context ~env ~curdir:dir ~objdir:dir
	      ~override:[ "PKGTOOLS_VERSION", pkgtools_version();
			]
	      ~make_cmd:"godi_make"
	      ~sys_mk_path [] 
	      ?stderr:protocol
  in
  try
    Godi_make.fast_parse_eval pcache makefile ctx;
    List.map
      (fun varname ->
	 let loc = { Godi_make.loc_filename = "<Expansion of " ^ varname ^ ">";
		     Godi_make.loc_linenumber = 0 } in
	 Godi_make.expand_var ctx loc varname
      )
      variables
  with
      ( Godi_make.Parse_error(_,_)
      | Godi_make.Eval_error(_,_)
      | Godi_make.File_not_found _ as error
      ) ->
	print_error (error_msg error);
	raise General_error
;;

let query_makefile =
  (* classic_query_makefile *)
  modern_query_makefile ;;


let read_confopts dir =
  let fn = Filename.concat dir "CONFOPTS" in
  if Sys.file_exists fn then
    let c = read_file fn in
    Str.split sloppy_lines_re c
  else
    []
;;


module VId = struct
  type t = var_id
  let compare = (compare : t -> t -> int)
end;;

module VIdMap = Map.Make(VId);;


class source_package_impl ?protocol (dir:string) : source_package =
  let confopts = read_confopts dir in
  let values =
    query_makefile
      ?protocol
      dir (["PKGNAME"; "COMMENT"; "DEPENDS"; "BUILD_DEPENDS";
	    "SELF_UPDATE_PKG";
	   ] @ confopts) in

  let (fullname,comment,depends,build_depends,self_upd_pkg,confopts_vals) =
    match values with
	fullname :: comment :: depends :: build_depends :: self_upd_pkg ::
	confopts_vals ->
	    (fullname, comment, depends, build_depends, self_upd_pkg,
	     confopts_vals)
      | _ ->
	  assert false
  in

  let (name, version, pkgrevision) = split_name_version fullname in
  let init_dependencies = parse_dep_strings depends in
  let init_build_dependencies = parse_dep_strings build_depends in

object(self)
  val mutable user_status = `Keep
  val mutable user_options = None

  val dependencies = init_dependencies
  val build_dependencies = init_build_dependencies

  val mutable config_dialog_shown = false

  val var_id = new_var_id()
  val mutable variants = VIdMap.empty

  initializer
    variants <- VIdMap.add var_id self variants

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description =
    read_optional_file (Filename.concat dir "DESCR")

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method var_id = var_id

  method self_update_pkg = self_upd_pkg
  method variant vid =
    try
      VIdMap.find vid variants
    with
	Not_found ->
	  failwith "class source_package: No such variant"

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Build | `Build_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method source_directory = dir

  method build_message =
    read_optional_file (Filename.concat dir "BUILDMSG")

  method build_options =
    confopts

  method values_of_build_options =
    List.combine confopts confopts_vals

  method values_of_user_build_options =
    match user_options with
	Some opts -> opts
      | None -> self#values_of_build_options

  method set_user_build_options opts =
    let names = List.map fst opts in
    if names <> self#build_options then
      invalid_arg "set_user_build_options";

    user_options <- Some opts

  method config_dialog_shown = config_dialog_shown

  method set_config_dialog_shown b = config_dialog_shown <- b

  method rescan_deps ?protocol godi_conf_file vid =
    let values =
      query_makefile
	?protocol
	~godi_conf:godi_conf_file
	dir [ "DEPENDS"; "BUILD_DEPENDS"] in

    let v_deps, v_bld_deps =
      match values with
	  [ v_deps; v_bld_deps ] -> (v_deps, v_bld_deps)
	| _ -> assert false in

    let new_pkg =
      {< dependencies = parse_dep_strings v_deps;
	 build_dependencies = parse_dep_strings v_bld_deps;
	 var_id = vid
      >} in

    variants <- VIdMap.add vid new_pkg variants;
    ()

end
;;


exception Next of string list
;;

let list_subdirs dirname =
  (* Return the list of all subdirectories of [dirname]; the returned names
   * include [dirname].
   *)
  (* TODO: Base implementation on Godi_file *)
  let rec read d l =
    try
      let x = Unix.readdir d in
      if x = "." || x = ".." then raise(Next []);
      let n = Filename.concat dirname x in
      let stat = Unix.stat n in
      if stat.Unix.st_kind = Unix.S_DIR then
	raise(Next [n])
      else
	raise(Next [])
    with
	End_of_file ->
	  l
      | Next xl ->
	  read d (xl @ l)
      | Unix.Unix_error(Unix.ENOENT,_,_) ->  (* e.g. bad symlink *)
	  read d l
  in
  let d = Unix.opendir dirname in
  try
    let l = read d [] in
    Unix.closedir d;
    l
  with
      error -> Unix.closedir d; raise error
;;


let really_scan_source_packages ?protocol () =
  (* To get a list of all source packages, traverse the subdirectories
   * at LOCALBASE/build. All directories two levels below this location
   * that have a "Makefile" and a "DESCR" file are considered as
   * source package.
   * This is a lousy definition!
   *)
  let build = Filename.concat (localbase()) "build" in
  let level1 =
    list_subdirs build in
  let level2 =
    List.flatten
      (List.map (fun dirname -> list_subdirs dirname) level1) in
  let source_dirs =
    List.filter
      (fun dirname ->
	 Sys.file_exists(Filename.concat dirname "Makefile") &&
	 Sys.file_exists(Filename.concat dirname "DESCR"))
      level2 in


  List.flatten
    (List.map
       (fun dir ->
	  try
	    let src_pkg = new source_package_impl ?protocol dir in
	    [ `Source_pkg src_pkg ]
	  with
	      error ->
		to_prot protocol
		  (sprintf
		     "Scan of directory %s: %s\n" dir (error_msg error));
		[]
       )
       source_dirs)
;;


let source_packages = ref None;;


let scan_source_packages ?(rescan=false) ?protocol () =
  if rescan then source_packages := None;
  match !source_packages with
      None ->
	let pkg_list = really_scan_source_packages ?protocol () in
	source_packages := Some pkg_list;
	pkg_list
    | Some pkg_list ->
	pkg_list
;;


let forget_source_packages() =
  source_packages := None
;;


(***********************************************************************)
(* Installed packages                                                  *)
(***********************************************************************)

let name_re = Str.regexp "^@name[ \t]+\\([^ \t]+\\)[ \t]*$"
let pkgdep_re = Str.regexp "^@pkgdep[ \t]+\\([^ \t]+\\)[ \t]*$"
let blddep_re = Str.regexp "^@blddep[ \t]+\\([^ \t]+\\)[ \t]*$"


let parse_CONTENTS dir =
  (* Extracts the name, the dependencies, and the build dependencies *)
  let filename = Filename.concat dir "+CONTENTS" in
  let name = ref "" in
  let pkgdeps = ref [] in
  let blddeps = ref [] in

  let file = open_in filename in
  try
    while true do
      let line = input_line file in
      if Str.string_match name_re line 0 then (
	name := Str.matched_group 1 line
      )
      else if Str.string_match pkgdep_re line 0 then (
	pkgdeps := Str.matched_group 1 line :: !pkgdeps
      )
      else if Str.string_match blddep_re line 0 then (
	blddeps := Str.matched_group 1 line :: !blddeps
      )
	(* else: ignore line *)
    done;
    assert false
  with
      End_of_file ->
	close_in file;
	( !name,
	  List.map parse_dependency !pkgdeps,
	  List.map parse_dependency !blddeps
	)
    | error ->
	close_in file;
	raise error
;;


class installed_package_impl (dir:string) : installed_package =

  let (fullname, dependencies, build_dependencies) =
    parse_CONTENTS dir in

  let comment = read_file (Filename.concat dir "+COMMENT") in

  let preserve = Sys.file_exists (Filename.concat dir "+PRESERVE") in

  let (name, version, pkgrevision) = split_name_version fullname in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = read_file (Filename.concat dir "+DESC")

  method preserve = preserve

  method dependencies = dependencies
  method build_dependencies = build_dependencies

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Remove | `Remove_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class installed_package: No such variant"

end
;;


let really_scan_installed_packages ?protocol () =
  (* To get a list of all source packages, traverse the subdirectories
   * at PKG_DBDIR. All subdirectories that have a "+CONTENTS" file
   * are considered as installed package.
   *)
  let dbdir = pkgdbdir() in
  let level1 =
    list_subdirs dbdir in
  let inst_dirs =
    List.filter
      (fun dirname ->
	 Sys.file_exists(Filename.concat dirname "+CONTENTS"))
      level1 in

  List.flatten
    (List.map
       (fun dir ->
	  try
	    let inst_pkg = new installed_package_impl dir in
	    [ `Installed_pkg inst_pkg ]
	  with
	      error ->
		to_prot protocol
		  (sprintf
		     "Scan of directory %s: %s\n" dir (error_msg error));
		[]
       )
       inst_dirs)
;;


let installed_packages = ref None;;


let scan_installed_packages ?(rescan=false) ?protocol () =
  if rescan then installed_packages := None;
  match !installed_packages with
      None ->
	let pkg_list = really_scan_installed_packages ?protocol () in
	installed_packages := Some pkg_list;
	pkg_list
    | Some pkg_list ->
	pkg_list
;;


let forget_installed_packages() =
  installed_packages := None
;;



(***********************************************************************)


let matches_dep pkg (name,cond) =
  let p = any_pkg pkg in
  p#name = name && version_matches_condition p#version cond
;;
