/* $Id: term_stubs.c 63 2003-08-24 15:39:25Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 */

/* Terminfo / Termcap access for fundamental terminal capabilities */

#include <stdlib.h>
#include <stdio.h>

#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/callback.h>
#include <caml/custom.h>
#include <caml/intext.h>

#include "term_config.h"


#ifdef HAVE_TERMINFO
#include <curses.h>
#include <term.h>
#else
#ifdef HAVE_TERMCAP
#include <termcap.h>
static char *termcap_bp;
#endif
#endif

CAMLprim value mlterm_setup(value ignore) {
#ifdef HAVE_TERMINFO
    setupterm(NULL, 1, NULL);
#else
#ifdef HAVE_TERMCAP
    char *termname;
    int rc;
    termcap_bp = malloc(1024);
    termname = getenv("TERM");
    if (termname == NULL) {
	termname = "dumb";   /* default */
    };
    rc = tgetent(termcap_bp, termname);
    if (rc == -1) {
	failwith("Cannot open termcap database");
    };
    if (rc == 0) {
	failwith("Terminal entry not found");
    };
#else
    /* do nothing */
#endif
#endif
    return Val_unit;
}

CAMLprim value mlterm_lines(value ignore) {
    int r;
    char *s;
    r = (-1);
    s = getenv("LINES");
    if (s != NULL) {
	r = atoi(s);
    }
#ifdef HAVE_TERMINFO
    if (r < 0) {
	r = tigetnum("lines");
	if (r < 0) failwith("Cannot query the terminfo capability 'lines'");
    }
#else
#ifdef HAVE_TERMCAP
    if (r < 0) {
	r = tgetnum("li");
	if (r < 0) failwith("Cannot query the termcap capability 'li'");
    }
#else
    if (r < 0) {
	r = 24;
    }
#endif
#endif
    return Val_int(r);
}

CAMLprim value mlterm_columns(value ignore) {
    int r;
    char *s;
    r = (-1);
    s = getenv("COLUMNS");
    if (s != NULL) {
	r = atoi(s);
    }
#ifdef HAVE_TERMINFO
    if (r < 0) {
	r = tigetnum("cols");
	if (r < 0) failwith("Cannot query the terminfo capability 'cols'");
    }
#else
#ifdef HAVE_TERMCAP
    if (r < 0) {
	r = tgetnum("co");
	if (r < 0) failwith("Cannot query the termcap capability 'co'");
    }
#else
    if (r < 0) {
	r = 80;
    }
#endif
#endif
    return Val_int(r);
}

CAMLprim value mlterm_getflag(value capname, value infoname) {
    int r;
#ifdef HAVE_TERMINFO
    r = tigetflag(String_val(infoname));
    if (r < 0) raise_not_found();
#else
#ifdef HAVE_TERMCAP
    r = tgetflag(String_val(capname));
#else
    raise_not_found();
#endif
#endif
    return Val_bool(r);
}

CAMLprim value mlterm_getnum(value capname, value infoname) {
    int r;
#ifdef HAVE_TERMINFO
    r = tigetnum(String_val(infoname));
    if (r < 0) raise_not_found();
#else
#ifdef HAVE_TERMCAP
    r = tgetnum(String_val(capname));
    if (r < 0) raise_not_found();
#else
    raise_not_found();
#endif
#endif
    return Val_int(r);
}

CAMLprim value mlterm_getstr(value capname, value infoname) {
    char *r;
#ifdef HAVE_TERMINFO
    r = tigetstr(String_val(infoname));
    if (r == (char *) 0 || r == (char *) (-1)) raise_not_found();
#else
#ifdef HAVE_TERMCAP
    char area[1025];
    char *areapnt;
    char *c;
    areapnt = area;
    c = tgetstr(String_val(capname), &areapnt);
    if (c == NULL) raise_not_found();
    if (areapnt - area >= 1024) {
	/* Normally crashes before we get here... */
	fprintf(stderr, "Termcap capability too long!");
	exit(99);   /* very bad... */
    }
    r = c;
#else
    raise_not_found();
#endif
#endif
    return copy_string(r);
}


static char *buffer = NULL;
static int   buflen;
static int   bufmaxlen;


static int add_char(int c) {
    if (buffer == NULL) {
	buffer = malloc(1024);
	bufmaxlen = 1024;
	buflen = 0;
    }
    if (buflen >= bufmaxlen) {
	char *newbuf;
	bufmaxlen = bufmaxlen * 2;
	newbuf = malloc(bufmaxlen);
	memcpy(newbuf, buffer, buflen);
	free(buffer);
	buffer = newbuf;
    }
    buffer[buflen] = c;
    buflen++;
    return c;
}


CAMLprim value mlterm_convertstr(value str) {
    value r;
#ifdef HAVE_TERMINFO
    buflen = 0;
    tputs(String_val(str), 1, add_char);
    add_char(0);
    r = copy_string(buffer);
#else
#ifdef HAVE_TERMCAP
    buflen = 0;
    tputs(String_val(str), 1, add_char);
    add_char(0);
    r = copy_string(buffer);
#else
    r = str;
#endif
#endif
    return r;
}
